@@ -1,14 +1,16 @@
+/* $Id: CoinMessageHandler.cpp 1854 2015-12-18 14:18:54Z forrest $ */
 // Copyright (C) 2002, International Business Machines
 // Corporation and others.  All Rights Reserved.
-
-#include "CoinFinite.hpp"
+// This code is licensed under the terms of the Eclipse Public License (EPL).
 
 #include "CoinMessageHandler.hpp"
 #include "CoinHelperFunctions.hpp"
 #include <cassert>
 #include <cstdlib>
+#include <cstddef>
 #include <map>
 
+
 /* Default constructor. */
 CoinOneMessage::CoinOneMessage()
 {
@@ -115,7 +117,7 @@ CoinMessages::CoinMessages(const CoinMessages & rhs)
   } else {
     char * temp = CoinCopyOfArray(reinterpret_cast<char *> (rhs.message_),lengthMessages_);
     message_ = reinterpret_cast<CoinOneMessage **> (temp);
-    long int offset = temp - reinterpret_cast<char *> (rhs.message_);
+    std::ptrdiff_t offset = temp - reinterpret_cast<char *> (rhs.message_);
     int i;
     //printf("new address %x(%x), rhs %x - length %d\n",message_,temp,rhs.message_,lengthMessages_);
     for (i=0;i<numberMessages_;i++) {
@@ -160,7 +162,7 @@ CoinMessages::operator=(const CoinMessages & rhs)
     } else {
       char * temp = CoinCopyOfArray(reinterpret_cast<char *> (rhs.message_),lengthMessages_);
       message_ = reinterpret_cast<CoinOneMessage **> (temp);
-      long int offset = temp - reinterpret_cast<char *> (rhs.message_);
+      std::ptrdiff_t offset = temp - reinterpret_cast<char *> (rhs.message_);
       int i;
       //printf("new address %x(%x), rhs %x - length %d\n",message_,temp,rhs.message_,lengthMessages_);
       for (i=0;i<numberMessages_;i++) {
@@ -279,14 +281,14 @@ void
 CoinMessages::toCompact()
 {
   if (numberMessages_&&lengthMessages_<0) {
-    lengthMessages_=numberMessages_*sizeof(CoinOneMessage *);
+    lengthMessages_=numberMessages_*CoinSizeofAsInt(CoinOneMessage *);
     int i;
     for (i=0;i<numberMessages_;i++) {
       if (message_[i]) {
-	int length = strlen(message_[i]->message());
-	length = (message_[i]->message()+length+1)-
-	  reinterpret_cast<char *> (message_[i]);
-	assert (length<1000);
+	int length = static_cast<int>(strlen(message_[i]->message()));
+	length = static_cast<int>((message_[i]->message()+length+1)-
+				      reinterpret_cast<char *> (message_[i]));
+	assert (length<COIN_MESSAGE_HANDLER_MAX_BUFFER_SIZE);
 	int leftOver = length %8;
 	if (leftOver)
 	  length += 8-leftOver;
@@ -296,17 +298,17 @@ CoinMessages::toCompact()
     // space
     char * temp = new char [lengthMessages_];
     CoinOneMessage ** newMessage = reinterpret_cast<CoinOneMessage **> (temp);
-    temp += numberMessages_*sizeof(CoinOneMessage *);
+    temp += numberMessages_*CoinSizeofAsInt(CoinOneMessage *);
     CoinOneMessage message;
     //printf("new address %x(%x) - length %d\n",newMessage,temp,lengthMessages_);
-    lengthMessages_=numberMessages_*sizeof(CoinOneMessage *);
+    lengthMessages_=numberMessages_*CoinSizeofAsInt(CoinOneMessage *);
     for (i=0;i<numberMessages_;i++) {
       if (message_[i]) {
 	message = *message_[i];
-	int length = strlen(message.message());
-	length = (message.message()+length+1)-
-	  reinterpret_cast<char *> (&message);
-	assert (length<1000);
+	int length = static_cast<int>(strlen(message.message()));
+	length = static_cast<int>((message.message()+length+1)-
+				  reinterpret_cast<char *> (&message));
+	assert (length<COIN_MESSAGE_HANDLER_MAX_BUFFER_SIZE);
 	int leftOver = length %8;
 	memcpy(temp,&message,length);
 	newMessage[i]=reinterpret_cast<CoinOneMessage *> (temp);
@@ -369,11 +371,22 @@ CoinMessageHandler::internalPrint()
   }
   return returnCode;
 }
+#if FLUSH_PRINT_BUFFER >= 2
+extern int coinFlushBufferFlag;
+#endif
 // Print message, return 0 normally
 int 
 CoinMessageHandler::print() 
 {
   fprintf(fp_,"%s\n",messageBuffer_);
+#if FLUSH_PRINT_BUFFER 
+#if FLUSH_PRINT_BUFFER < 2
+  fflush(fp_);
+#else
+  if (coinFlushBufferFlag)
+    fflush(fp_);
+#endif
+#endif
   return 0;
 }
 // Check severity
@@ -396,20 +409,47 @@ CoinMessageHandler::checkSeverity()
    printf messages in code
 */
 void 
-CoinMessageHandler::setLogLevel(int value)
+CoinMessageHandler::setLogLevel (int value)
 {
-  if (value>=-1)
-    logLevel_=value;
+  if (value >= -1) logLevel_ = value ;
 }
 void 
-CoinMessageHandler::setLogLevel(int which,int value)
+CoinMessageHandler::setLogLevel (int which, int value)
 {
-  if (which>=0&&which<COIN_NUM_LOG) {
-    if (value>=-1)
-      logLevel_=value;
+  if (which >= 0 && which < COIN_NUM_LOG) {
+    if (value >= -1)
+      logLevels_[which] = value ;
   }
 }
 void 
+CoinMessageHandler::setPrecision(unsigned int new_precision) {
+
+  char new_string[8] = {'%','.','8','f','\0','\0','\0','\0'};
+  
+  //we assume that the precision is smaller than one thousand
+  new_precision = std::min<unsigned int>(999,new_precision);
+  if (new_precision == 0)
+    new_precision = 1;
+  g_precision_ = new_precision ;
+  int idx = 2;
+  int base = 100;
+  bool print = false;
+  while (base > 0) {
+
+    char c = static_cast<char>(new_precision / base);
+    new_precision = new_precision % base;
+    if (c != 0)
+      print = true;
+    if (print) {
+      new_string[idx] = static_cast<char>(c +  '0');
+      idx++;
+    }
+    base /= 10;
+  }
+  new_string[idx] = 'g';
+  strcpy(g_format_,new_string);
+}
+void 
 CoinMessageHandler::setPrefix(bool value)
 {
   if (value)
@@ -433,8 +473,13 @@ CoinMessageHandler::CoinMessageHandler() :
   highestNumber_(-1),
   fp_(stdout)
 {
+  const char* g_default = "%.8g";
+
+  strcpy(g_format_,g_default);
+  g_precision_ = 8 ;
+  
   for (int i=0;i<COIN_NUM_LOG;i++)
-    logLevels_[i]=1;
+    logLevels_[i]=-1000;
   messageBuffer_[0]='\0';
   messageOut_ = messageBuffer_;
   source_="Unk";
@@ -450,8 +495,13 @@ CoinMessageHandler::CoinMessageHandler(FILE * fp) :
   highestNumber_(-1),
   fp_(fp)
 {
+  const char* g_default = "%.8g";
+  
+  strcpy(g_format_,g_default);
+  g_precision_ = 8 ;
+
   for (int i=0;i<COIN_NUM_LOG;i++)
-    logLevels_[i]=1;
+    logLevels_[i]=-1000;
   messageBuffer_[0]='\0';
   messageOut_ = messageBuffer_;
   source_="Unk";
@@ -480,7 +530,7 @@ CoinMessageHandler::gutsOfCopy(const CoinMessageHandler& rhs)
   longValue_=rhs.longValue_;
   charValue_=rhs.charValue_;
   stringValue_=rhs.stringValue_;
-  long int offset ;
+  std::ptrdiff_t offset ;
   if (rhs.format_)
   { offset = rhs.format_ - rhs.currentMessage_.message();
     format_ = currentMessage_.message()+offset; }
@@ -494,6 +544,8 @@ CoinMessageHandler::gutsOfCopy(const CoinMessageHandler& rhs)
   highestNumber_= rhs.highestNumber_;
   fp_ = rhs.fp_;
   source_ = rhs.source_;
+  strcpy(g_format_,rhs.g_format_);
+  g_precision_ = rhs.g_precision_ ;
 }
 /* The copy constructor */
 CoinMessageHandler::CoinMessageHandler(const CoinMessageHandler& rhs)
@@ -515,78 +567,135 @@ CoinMessageHandler::clone() const
 {
   return new CoinMessageHandler(*this);
 }
-// Start a message
-CoinMessageHandler & 
-CoinMessageHandler::message(int messageNumber,
-			   const CoinMessages &normalMessage)
+
+/*
+  Decide if we're printing or not. Split out because it's nontrivial and
+  used in two places.
+
+  Recall that setLogLevel(lvl) sets logLevel_. To set logLevels_[i], use
+  setLogLevel(i,lvl). The two are not connected. By default, all entries
+  of logLevels_ are initialised to -1000.
+
+  Unless the client changes logLevels_[0], when the log level (lvl)
+  of the message is 8 or more, it's ANDed with the handler's log level
+  (logLevel_). Printing is enabled only if the result is nonzero. Commonly
+  this is used in a scheme where individual bits enable particular debug
+  output.
+*/
+void CoinMessageHandler::calcPrintStatus (int msglvl, int msgclass)
 {
-  if (messageOut_!=messageBuffer_) {
-    // put out last message
-    internalPrint();
-  }
-  internalNumber_=messageNumber;
-  currentMessage_= *(normalMessage.message_[messageNumber]);
-  source_ = normalMessage.source_;
-  format_ = currentMessage_.message_;
-  messageBuffer_[0]='\0';
-  messageOut_=messageBuffer_;
-  highestNumber_ = CoinMax(highestNumber_,currentMessage_.externalNumber_);
-  // do we print
-  int detail = currentMessage_.detail_;
-  printStatus_=0;
-  if (detail>=8&&logLevel_>=0) {
-    // bit setting - debug
-    if ((detail&logLevel_)==0)
-      printStatus_ = 3;
-  } else if (logLevel_<detail) {
+  printStatus_ = 0 ;
+  if (logLevels_[0] == -1000) {
+    if (msglvl >= 8 && logLevel_ >= 0) {
+      if ((msglvl&logLevel_) == 0)
+	printStatus_ = 3 ;
+    } else if (logLevel_ < msglvl) {
+      printStatus_ = 3 ;
+    }
+  } else if (logLevels_[msgclass] < msglvl) {
     printStatus_ = 3;
   }
+}
+
+/*
+  Start a message using a standard CoinOneMessage.
+*/
+CoinMessageHandler & 
+CoinMessageHandler::message (int messageNumber,
+			     const CoinMessages &normalMessages)
+{
+  // Deal with the previous message, if there is one.
+  if (messageOut_ != messageBuffer_) {
+    internalPrint() ;
+  }
+  // Acquire the new message
+  internalNumber_ = messageNumber ;
+  currentMessage_ = *(normalMessages.message_[messageNumber]) ;
+  source_ = normalMessages.source_ ;
+  format_ = currentMessage_.message_ ;
+  highestNumber_ = CoinMax(highestNumber_,currentMessage_.externalNumber_);
+
+  // Initialise the message construction buffer
+  messageBuffer_[0] = '\0' ;
+  messageOut_ = messageBuffer_ ;
+
+  // Decide whether or not to print (sets printStatus_)
+  calcPrintStatus(currentMessage_.detail_,normalMessages.class_) ;
+
+  // If we're printing, initialise the message
   if (!printStatus_) {
     if (prefix_) {
       sprintf(messageOut_,"%s%4.4d%c ",source_.c_str(),
 	      currentMessage_.externalNumber_,
-	      currentMessage_.severity_);
-      messageOut_ += strlen(messageOut_);
+	      currentMessage_.severity_) ;
+      messageOut_ += strlen(messageOut_) ;
     }
-    format_ = nextPerCent(format_,true);
+    format_ = nextPerCent(format_,true) ;
   }
-  return *this;
+  return (*this) ;
 }
-/* The following is to help existing codes interface
-   Starts message giving number and complete text
+/*
+  Start a message, providing the full message, information to generate
+  a prefix, a severity code, and an optional log level.
+  Intended as an aid to help existing codes interface.
 */
 CoinMessageHandler & 
-CoinMessageHandler::message(int externalNumber,const char * source,
-			   const char * msg, char severity)
+CoinMessageHandler::message (int externalNumber, const char *source,
+			     const char *msg, char severity, int loglvl)
 {
-  if (messageOut_!=messageBuffer_) {
-    // put out last message
-    internalPrint();
+  // Deal with the previous message, if there is one.
+  if (messageOut_ != messageBuffer_) {
+    internalPrint() ;
   }
-  internalNumber_=externalNumber;
-  currentMessage_= CoinOneMessage();
-  currentMessage_.setExternalNumber(externalNumber);
-  source_ = source;
-  // mark so will not update buffer
-  printStatus_=2;
+  // Set up a dummy message.
+  internalNumber_ = externalNumber ;
+  char detail = ((loglvl >= 0)?(static_cast<char>(loglvl)):'\000') ;
+  currentMessage_= CoinOneMessage(externalNumber,detail,msg) ;
+  source_ = source ;
   highestNumber_ = CoinMax(highestNumber_,externalNumber);
-  // If we get here we always print
-  if (prefix_) {
-    sprintf(messageOut_,"%s%4.4d%c ",source_.c_str(),
-	    externalNumber,
-	    severity);
+
+  // Initialise the message construction buffer
+  messageBuffer_[0] = '\0' ;
+  messageOut_ = messageBuffer_ ;
+
+  /*
+    Decide whether or not to print. The normal value of printStatus_ here
+    is 2 (complete message provided). loglvl defaults to -1 for backwards
+    compatibility (previously there was no provision for a log level).
+  */
+  if (loglvl >= 0) calcPrintStatus(loglvl,0) ;
+  if (!printStatus_) {
+    printStatus_ = 2 ;
+    if (prefix_) {
+      sprintf(messageOut_,"%s%4.4d%c ",source_.c_str(),
+	      externalNumber,severity) ;
+    }
+    strcat(messageBuffer_,msg) ;
+    messageOut_ = messageBuffer_+strlen(messageBuffer_) ;
   }
-  strcat(messageBuffer_,msg);
-  messageOut_=messageBuffer_+strlen(messageBuffer_);
-  return *this;
+  return (*this) ;
 }
-  /* Allows for skipping printing of part of message,
-      but putting in data */
-  CoinMessageHandler & 
+
+/*
+  Decides whether or not to print and returns a reference to the handler.
+*/
+CoinMessageHandler & 
+CoinMessageHandler::message(int loglvl)
+{
+  // Adjust print status?
+  if (loglvl >= 0) calcPrintStatus(loglvl,0) ;
+
+  return (*this) ;
+}
+
+/*
+  Allows for skipping printing of part of message, but putting in data
+*/
+CoinMessageHandler & 
 CoinMessageHandler::printing(bool onOff)
 {
   // has no effect if skipping or whole message in
-  if (printStatus_<2) {
+  if (printStatus_ < 2) {
     assert(format_[1]=='?');
     *format_ = '%' ;
     if (onOff)
@@ -597,22 +706,32 @@ CoinMessageHandler::printing(bool onOff)
   }
   return *this;
 }
-/* Stop (and print) 
+/*
+  Stop (and print)
+
+  Unless printing is currently suppressed. We need to do the finishing actions
+  in any event.
 */
 int 
 CoinMessageHandler::finish()
 {
-  if (messageOut_!=messageBuffer_) {
-    // put out last message
+  // Deal with the collected message
+  if (printStatus_ < 3 && messageOut_ != messageBuffer_) {
     internalPrint();
   }
-  internalNumber_=-1;
-  format_ = NULL;
-  messageBuffer_[0]='\0';
-  messageOut_=messageBuffer_;
-  printStatus_=0;
-  return 0;
+  // Clean up for the next message.
+  internalNumber_ = -1 ;
+  format_ = NULL ;
+  messageBuffer_[0] = '\0' ;
+  messageOut_ = messageBuffer_ ;
+  printStatus_ = 0 ;
+  doubleValue_.clear() ;
+  longValue_.clear() ;
+  charValue_.clear() ;
+  stringValue_.clear() ;
+  return (0) ;
 }
+
 /* Gets position of next field in format
    If we're scanning the initial portion of the string (prior to the first
    `%' code) the prefix will be copied to the output buffer. Normally, the
@@ -628,7 +747,7 @@ CoinMessageHandler::nextPerCent(char * start , const bool initial)
       char * nextPerCent = strchr(start,'%');
       if (nextPerCent) {
 	if (initial&&!printStatus_) {
-	  int numberToCopy=nextPerCent-start;
+	  int numberToCopy=static_cast<int>(nextPerCent-start);
 	  strncpy(messageOut_,start,numberToCopy);
 	  messageOut_+=numberToCopy;
 	} 
@@ -695,19 +814,33 @@ CoinMessageHandler::operator<< (double doublevalue)
   if (printStatus_==3)
     return *this; // not doing this message
   doubleValue_.push_back(doublevalue);
+
   if (printStatus_<2) {
     if (format_) {
-      //format is at % (but changed to 0)
+      //format is at \0 (but changed to %)
       *format_='%';
       char * next = nextPerCent(format_+1);
       // could check
       if (!printStatus_) {
-	sprintf(messageOut_,format_,doublevalue);
+	if (format_[1] == '.' && format_[2] >= '0' && format_[2] <= '9') {
+	  // an explicitly specified precision currently overrides the
+	  // precision of the message handler
+	  sprintf(messageOut_,format_,doublevalue);
+	}
+	else {
+	  sprintf(messageOut_,g_format_,doublevalue);
+	  if (next != format_+2) {
+	    messageOut_+=strlen(messageOut_);
+	    sprintf(messageOut_,format_+2);
+	  }
+	}
 	messageOut_+=strlen(messageOut_);
       }
       format_=next;
     } else {
-      sprintf(messageOut_," %g",doublevalue);
+      sprintf(messageOut_," ");
+      messageOut_ += 1;
+      sprintf(messageOut_,g_format_,doublevalue);
       messageOut_+=strlen(messageOut_);
     } 
   }
@@ -794,7 +927,7 @@ CoinMessageHandler::operator<< (char charvalue)
 {
   if (printStatus_==3)
     return *this; // not doing this message
-  longValue_.push_back(charvalue);
+  charValue_.push_back(charvalue);
   if (printStatus_<2) {
     if (format_) {
       //format is at % (but changed to 0)
@@ -837,29 +970,26 @@ CoinMessageHandler::operator<< (const char *stringvalue)
   }
   return *this;
 }
+
+/*
+  Handle markers. Even when printing is suppressed (printStatus_ == 3) we need
+  to execute finish() to reset for the next message.
+*/
 CoinMessageHandler & 
 CoinMessageHandler::operator<< (CoinMessageMarker marker)
 {
-  if (printStatus_!=3) {
-    switch (marker) {
-    case CoinMessageEol:
-      finish();
-      break;
-    case CoinMessageNewline:
-      strcat(messageOut_,"\n");
-      messageOut_++;
-      break;
+  switch (marker) {
+    case CoinMessageEol: {
+      finish() ;
+      break ;
+    }
+    case CoinMessageNewline: {
+      if (printStatus_ != 3) {
+	strcat(messageOut_,"\n") ;
+	messageOut_++ ;
+      }
+      break ;
     }
-  } else {
-    // skipping - tidy up
-    format_ = NULL;
   }
-  return *this;
-}
-
-// returns current 
-CoinMessageHandler & 
-CoinMessageHandler::message()
-{
-  return * this;
+  return (*this) ;
 }
